from .util import list_get


class Option:
    """A choice option for a ChoiceInput.

    Attributes:
        value: The option's value.
        other: Whether or not this option is the "Other" option.
    """

    class _Index:
        VALUE = 0
        ACTION = 2
        OTHER = 4
        IMAGE = 5  # not implemented

    @classmethod
    def parse(cls, option):
        """Creates an Option from its JSON representation.

        This method should not be called directly,, use options.parse instead.
        """
        return cls(**cls._parse(option))

    @classmethod
    def _parse(cls, option):
        return {
            'value': option[cls._Index.VALUE] or '',
            'other': bool(list_get(option, cls._Index.OTHER, False)),
        }

    def __init__(self, *, value, other):
        self.value = value
        self.other = other

    def to_str(self, indent=0, with_value=None):
        """Returns a text representation of the option.

        For args description, see Form.to_str.
        """
        if self.other:
            if with_value is not None:
                return f'Other: "{with_value}"'
            return 'Other'
        return self.value


class ActionOption(Option):
    """An option, which, when chosen, may lead to a different page.

    Attributes:
        next_page: The next page for this option.
    """

    @classmethod
    def _parse(cls, option):
        res = super()._parse(option)
        res.update({
            'action': option[cls._Index.ACTION],
        })
        return res

    def __init__(self, *, action, **kwargs):
        super().__init__(**kwargs)
        self._action = action
        self.next_page = None

    def to_str(self, indent=0, with_value=None):
        """See base class."""
        from .elements import Page
        s = super().to_str(indent, with_value)
        if self.next_page is Page.SUBMIT:
            return f'{s} -> Submit'
        if self.next_page is None:
            return f'{s} -> Ignored'
        return f'{s} -> Go to Page {self.next_page.index + 1}'

    def _resolve_action(self, next_page, mapping):
        from .elements import _Action
        if next_page is None:
            return
        if self._action == _Action.NEXT:
            self.next_page = next_page
        else:
            self.next_page = mapping[self._action]


def parse(option):
    """Creates an Option of the right type from its JSON representation."""
    if len(option) > Option._Index.ACTION and option[Option._Index.ACTION] is not None:
        return ActionOption.parse(option)
    return Option.parse(option)



import random
import re
from enum import Enum
from typing import Optional

SEP_WIDTH = 60  # implement dynamic width?
RADIO_SYMBOLS = ('◯', '⦿')
CHECKBOX_SYMBOLS = ('☐', '☒')


# Pattern from form submision page
EMAIL_REGEX = re.compile(
    r"^[+a-zA-Z0-9_.!#$%&'*\/=?^`{|}~-]+@([a-zA-Z0-9-]+\.)+[a-zA-Z0-9]{2,63}$"
)

# From https://gist.github.com/dperini/729294
URL_REGEX = re.compile(
    r'^(?:(?:(?:https?|ftp):)?\/\/)(?:\S+(?::\S*)?@)?(?:(?!(?:10|127)(?:\.\d{1,3}){3})'
    r'(?!(?:169\.254|192\.168)(?:\.\d{1,3}){2})(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})'
    r'(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}'
    r'(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-z0-9\u00a1-\uffff]'
    r'[a-z0-9\u00a1-\uffff_-]{0,62})?[a-z0-9\u00a1-\uffff]\.)+'
    r'(?:[a-z\u00a1-\uffff]{2,}\.?))(?::\d{2,5})?(?:[/?#]\S*)?$'
)


class DefaultEnum(Enum):
    @classmethod
    def _missing_(cls, value):
        return cls.__members__['UNKNOWN']  # is it possible to add a member when creating a class?


class ArgEnum(Enum):
    def __new__(cls, value, arg):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.arg = arg
        return obj


def list_get(lst, index, default=None):
    if len(lst) > index:
        return lst[index]
    return default


def page_separator(indent):
    return '=' * SEP_WIDTH


def elem_separator(indent):
    return '—' * (SEP_WIDTH - indent)


def add_indent(text, indent):
    if not indent:
        return text
    spaces = ' ' * indent
    return ''.join(spaces + line for line in text.splitlines(keepends=True))


def random_subset(a, *, prob=0.5, min_size: Optional[int] = None, max_size: Optional[int] = None):
    """Choose a random subset of a.

    Args:
        a: the input sequence
        prob: Probability to pick an element
            (when called without size constraints).
        min_size: Minimal size of the result.
        max_size: Maximal size of the result.
    Returns:
        A random subset of a
    """
    def subset(a, prob):
        res = []
        remaining = []
        for el in a:
            if random.random() < prob:
                res.append(el)
            else:
                remaining.append(el)
        return res

    if min_size is None:
        min_size = 0
    if max_size is None:
        max_size = len(a)
    min_size = max(min_size, 0)
    max_size = min(max_size, len(a))

    if min_size > max_size:
        raise ValueError(f'Invalid size range for random_subset: [{min_size}, {max_size}]')

    tmp = random.sample(a, max_size)
    return tmp[:min_size] + subset(tmp[min_size:], prob)  # tweak distribution?



from __future__ import annotations

from abc import ABC, abstractmethod

from enum import Enum, auto
from typing import Dict, List, Literal, Union, cast, Any, Optional, Tuple, TYPE_CHECKING

from .util import DefaultEnum, list_get
from .validators import Validator, GridValidator, TextValidator, GridTypes, NumberTypes
from .errors import DuplicateOther, EmptyOther, InvalidChoice, \
    ElementTypeError, ElementValueError, RequiredElement, RequiredRow, RowTypeError, \
    InvalidRowChoice, MisconfiguredElement
from .options import ActionOption, Option, parse as parse_option

if TYPE_CHECKING:
    from .form import Form


class Value(Enum):
    """A value which can be returned by a callback in Form.fill

    DEFAULT: Use the value returned by the default callback.
    EMPTY: Leave the element empty.
    UNCHANGED: Leave the element value unchanged.
    """

    DEFAULT = auto()
    EMPTY = auto()
    UNCHANGED = auto()


EmptyValue = Literal[Value.EMPTY]
TextValue = str
ChoiceValue = Union[str, Option]
MultiChoiceValue = Union[ChoiceValue, List[ChoiceValue]]

GridChoiceValue = List[
    Union[ChoiceValue, EmptyValue]
]

GridMultiChoiceValue = List[
    Union[MultiChoiceValue, EmptyValue]
]

GridValue = Union[GridChoiceValue, GridMultiChoiceValue]


class _Action:
    FIRST = -1
    NEXT = -2
    SUBMIT = -3


class Element:
    """An element of a form.

    Attributes:
        id: The element ID.
        name: The element name (may be None)
        description: The element description (may be None).
        type: The element type.
    """

    class Type(DefaultEnum):
        UNKNOWN = -1
        SHORT = 0
        PARAGRAPH = 1
        RADIO = 2
        DROPDOWN = 3
        CHECKBOXES = 4
        SCALE = 5
        COMMENT = 6
        GRID = 7
        PAGE = 8
        DATE = 9
        TIME = 10
        IMAGE = 11
        VIDEO = 12
        FILE_UPLOAD = 13

    class _Index:
        ID = 0
        NAME = 1
        DESCRIPTION = 2
        TYPE = 3

    @classmethod
    def parse(cls, elem):
        """Creates an Element from its JSON representation.

        This method should not be called directly,
        use gforms.elements.parse instead.
        """
        return cls(**cls._parse(elem))

    @classmethod
    def _parse(cls, elem):
        return {
            'id_': elem[cls._Index.ID],
            'name': elem[cls._Index.NAME],
            'description': elem[cls._Index.DESCRIPTION],  # may be None
            'type_': Element.Type(elem[cls._Index.TYPE]),
        }

    def __init__(self, *, id_, name, description, type_):
        self.id = id_
        self.name = name
        self.description = description
        self.type = type_
        self._form: Optional['Form'] = None

    def bind(self, form):
        self._form = form

    def to_str(self, indent=0, **kwargs):
        """Returns a text representation of the element.

        For args description, see Form.to_str.
        """
        s = f'{self._type_str()}'
        if self.name:
            s = f'{s}: {self.name}'
        if self.description:
            s = f'{s}\n{self.description}'
        return s

    def _type_str(self):
        return type(self).__name__


class InputElement(Element, ABC):
    """An element which can be filled.

    Attributes:
        required: Self-explanatory.
    """

    class _Index(Element._Index):
        ENTRIES = 4
        ENTRY_ID = 0
        REQUIRED = 2

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        entries = cls._get_entries(elem)
        res.update({
            'entry_ids': [e[cls._Index.ENTRY_ID] for e in entries],
            'required': bool(entries[0][cls._Index.REQUIRED]),  # same for all entries
        })
        return res

    def __init__(self, *, entry_ids, required, **kwargs):
        super().__init__(**kwargs)
        self._entry_ids = entry_ids
        self.required = required
        self._values: List[List[str]] = [[] for _ in self._entry_ids]

    @abstractmethod
    def set_value(self, value):
        """Sets the value for this element.

        All elements accept Value.EMPTY as an input.
        For other accepted types and values, see subclasses.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        raise NotImplementedError()

    def validate(self):
        """Checks if the element has a valid value.

        Raises a ValidationError if the value is invalid.

        Raises:
            gforms.errors.RequiredElement:
                This element is required, but the value is empty.
            gforms.errors.InvalidValue:
                The value has a correct type,
                but is not valid for this element.
        """
        self._validate()
        if self._form is not None:
            self._form._unvalidated_elements.discard(self)

    def to_str(self, indent=0, include_answer=False):
        """See base class."""
        parts = self._header()
        hints = self._hints(indent, include_answer)
        if hints:
            parts.append('\n')
            parts.append('\n'.join(hints))
        if include_answer:
            answer = self._answer()
            if answer:
                parts.append('\n')
                parts.append('\n'.join(answer))
        return ''.join(parts)

    def payload(self) -> Dict[str, List[str]]:
        """Returns a payload which can be used in Form.submit

        The payload contains the element's value(s)
        and can be used as a (part of) POST request body.
        """
        payload = {}
        for entry_id, value in zip(self._entry_ids, self._values):
            if value:
                payload[self._submit_id(entry_id)] = value[:]
        return payload

    def draft(self) -> List[Tuple]:
        """Returns a (part of) emulated draftResponse for Form.submit."""
        result = []
        for entry_id, value in zip(self._entry_ids, self._values):
            if value:
                result.append((
                    None,  # ???
                    entry_id,
                    value[:],
                    0  # Last choice is "Other" (for "Individual responses" tab)
                ))
        return result

    def prefill(self, prefilled_data: Dict[str, List[str]]):
        """Fills the element with values from the prefilled link."""
        values = [prefilled_data.get(entry_id, []) for entry_id in self._entry_ids]
        # prefilled data still needs to be validated
        # (e.g. a modified url was used or elements were updated)
        # Also, there may be no prefilled values for a required element.
        self._set_values(values)

    @staticmethod
    def _submit_id(entry_id):
        return f'entry.{entry_id}'

    @staticmethod
    def _get_entries(elem):
        return elem[InputElement._Index.ENTRIES]

    def _is_empty(self):
        return not any(self._values)

    def _set_values(self, values: List[Union[List[str], EmptyValue]]):
        values = values[:]
        for i, entry_values in enumerate(values):
            if entry_values is Value.EMPTY:
                values[i] = []

        self._values = values  # type: ignore
        if self._form is not None:
            self._form._unvalidated_elements.add(self)

    def _header(self) -> List[str]:
        parts = [self._type_str()]
        if self.required:
            parts.append('*')
        if self.name:
            parts.append(f': {self.name}')
        if self.description:
            parts.append(f'\n{self.description}')
        return parts

    def _hints(self, indent=0, modify=False):
        """Returns input hints (options / values).

        Returned strings should be already indented.
        """
        return []

    def _answer(self) -> List[str]:
        return [f'> {self._entry_answer(i)}' for i in range(len(self._entry_ids))]

    def _entry_answer(self, i) -> str:
        entry_val = self._values[i]
        if not entry_val:
            return 'EMPTY'
        return ', '.join(f'"{value}"' for value in entry_val)

    def _validate(self):
        for i in range(len(self._values)):
            self._validate_entry(i)

    def _validate_entry(self, index):
        if self.required and not self._values[index]:
            raise RequiredElement(self, index=index)


class SingleInput(InputElement):
    """An InputElement with a single entry."""

    def _set_value(self, value: Union[List[str], EmptyValue]):
        self._set_values([value])

    @property
    def _value(self):
        return self._values[0]

    @property
    def _entry_id(self):
        return self._entry_ids[0]

    def _part_id(self, name):
        return f'{self._submit_id(self._entry_id)}_{name}'

    @staticmethod
    def _get_entry(elem):
        return InputElement._get_entries(elem)[0]


class ChoiceInput(InputElement, ABC):
    """An input element which has a predefined set of options.

    Attributes:
        options: A list of allowed choices for this element.
    """

    class _Index(InputElement._Index):
        OPTIONS = 1

    # This class is basically the same as CheckboxGrid
    _has_multichoice = False

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        options = [
            [
                parse_option(opt) for opt in entry[cls._Index.OPTIONS]
            ] for entry in cls._get_entries(elem)
        ]
        res.update({
            'options': options,
        })
        return res

    def __init__(self, *, options: List[List[Option]], **kwargs):
        super().__init__(**kwargs)
        self._options = options

    @property
    @abstractmethod
    def options(self):
        raise NotImplementedError()

    def _set_choices(self, choices: List[Union[List[ChoiceValue], EmptyValue]]):
        new_choices: List[Union[List[str], EmptyValue]] = [[] for _ in choices]
        for i, entry_choices in enumerate(choices):
            if entry_choices is Value.EMPTY:
                continue
            for choice in cast(List[ChoiceValue], entry_choices):
                self._add_choice(
                    new_choices[i],
                    self._find_option(choice, i)
                )
        self._set_values(new_choices)

    def _hints(self, indent=0, modify=False):
        return []

    def _find_option(self, value: ChoiceValue, i):
        if isinstance(value, Option):
            # assuming value was chosen from self.options
            return value

        for opt in self._options[i]:
            if opt.value == value:
                return opt

        raise InvalidChoice(self, value, index=i)

    def _add_choice(self, choices, choice: Option):
        choices.append(choice.value)

    def _to_choice_list(self, choices: Union[MultiChoiceValue, EmptyValue, Any])\
            -> Union[List[ChoiceValue], EmptyValue]:
        if choices is Value.EMPTY:
            return choices
        elif isinstance(choices, list):
            if not self._has_multichoice:
                raise ElementTypeError(self, choices)
            for choice in choices:
                if not ChoiceInput._is_choice_value(choice):
                    raise ElementTypeError(self, choices)
            return choices
        elif ChoiceInput._is_choice_value(choices):
            return [choices]
        raise ElementTypeError(self, choices)

    @staticmethod
    def _is_choice_value(value):
        return isinstance(value, str) or isinstance(value, Option)


class MultiChoiceInput(ChoiceInput):
    """A ChoiceInput which can accept multiple values."""

    _has_multichoice = True


class ChoiceInput1D(ChoiceInput, SingleInput):
    """A ChoiceInput with a single entry.

    Attributes:
        shuffle_options: Self-explanatory..
    """

    # ChoiceInput1D == Checkboxes without "other"
    _choice_symbols = ('·', '>')

    class _Index(ChoiceInput._Index):
        SHUFFLE_OPTIONS = 8  # Does not affect Scale

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        res.update({
            'shuffle_options': bool(list_get(cls._get_entry(elem), cls._Index.SHUFFLE_OPTIONS, False)),
        })
        return res

    def __init__(self, *, shuffle_options, **kwargs):
        super().__init__(**kwargs)
        self.shuffle_options = shuffle_options


    @property
    def options(self):
        return self._options[0]

    def _hints(self, indent=0, modify=False):
        if modify:
            return [
                f'{self._choice_symbols[opt.value in self._value]} {opt.to_str()}'
                for opt in self.options
            ]
        else:
            return [f'{self._choice_symbols[0]} {opt.to_str()}' for opt in self.options]

    def _answer(self):
        return []


class OtherChoiceInput(ChoiceInput1D):
    """A ChoiceInput1D which may have an "Other" option.

    Attributes:
        other_option: The "Other" option, if it exists.
    """

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        options = []
        other_option = None
        for opt in res['options'][0]:
            if opt.other:
                other_option = opt
            else:
                options.append(opt)
        res.update({
            'options': [options],
            'other_option': other_option,
        })
        return res

    def __init__(self, *, other_option, **kwargs):
        super().__init__(**kwargs)
        self.other_option = other_option
        self._other_value: Union[str, None] = None

    def payload(self) -> Dict[str, List[str]]:
        """See base class."""
        payload = super().payload()
        # '' is not converted to None (see _validate_entry), so "... not None" isn't applicable
        if not self._other_value:
            return payload

        main_key = self._submit_id(self._entry_id)
        other_key = main_key + '.other_option_response'
        payload.setdefault(main_key, []).append('__other_option__')
        payload[other_key] = [self._other_value]
        return payload

    def draft(self) -> List[Tuple]:
        """See base class."""
        result = super().draft()
        if not self._other_value:
            return result

        if not result:
            result.append((None, self._entry_id, [self._other_value], 1))
            return result
        *data, answers, _ = result[0]
        answers.append(self._other_value)
        result[0] = (*data, answers, 1)
        return result

    def _is_empty(self):
        return super()._is_empty() and self._other_value is None

    def _set_choices(self, choices: List[Union[List[ChoiceValue], EmptyValue]]):
        self._other_value = None
        super()._set_choices(choices)

    def _find_option(self, value: ChoiceValue, i):  # i == 0
        if self.other_option is None:
            return super()._find_option(value, i)

        if isinstance(value, Option):
            if value.other and self._other_value is not None:
                raise DuplicateOther(self, self._other_value, value.value)
            return value

        for opt in self.options:
            if opt.value == value:
                return opt

        if self._other_value is not None:
            raise DuplicateOther(self, self._other_value, value)
        self.other_option.value = value
        return self.other_option

    def _add_choice(self, choices, choice: Option):
        if choice.other:
            self._other_value = choice.value
        else:
            super()._add_choice(choices, choice)

    def _validate_entry(self, index):
        if self.required and not self._value:
            if self._other_value is None:
                raise RequiredElement(self)
            if not self._other_value:
                raise EmptyOther(self)

    def _hints(self, indent=0, modify=False):
        hints = super()._hints(indent, modify)
        if self.other_option is not None:
            if modify and self._other_value is not None:
                hints.append(
                    f'{self._choice_symbols[1]}'
                    f' {self.other_option.to_str(with_value=self._other_value)}'
                )
            else:
                hints.append(f'{self._choice_symbols[0]} {self.other_option.to_str()}')
        return hints


class ValidatedInput(InputElement, ABC):
    """An input element which may have a validator.

    Attributes:
        validator: The validator, if it exists.
    """

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        res.update({
            'validator': cls._parse_validator(elem)
        })
        return res

    def __init__(self, *, validator, **kwargs):
        super().__init__(**kwargs)
        self.validator: Optional[Validator] = validator

    def is_misconfigured(self):
        """Checks if the element cannot be filled with a valid value."""
        if self.validator is None:
            return False
        if self.validator.has_unknown_type() or self.validator.bad_args:
            return False
        if not self.required:  # always can set empty value
            return False
        return self._is_misconfigured()

    def _validate(self):
        """See base validate method.

        If this element is required and MisconfiguredElement is raised,
        you should skip the page with this element, if possible.
        Otherwise, the form is not submittable.

        Raises:
            gforms.errors.MisconfiguredElement:
                This element will not accept any non-empty value.
        """
        if self.is_misconfigured():
            raise MisconfiguredElement(self)
        super()._validate()
        if self.validator is not None and not self._is_empty():  # Non-empty value
            self.validator.validate(self)

    def _hints(self, indent=0, modify=False):
        res = []
        if self.validator is not None:
            res.append(f'! {self.validator.to_str()} !')
        return res + super()._hints(indent, modify)

    @classmethod
    @abstractmethod
    def _parse_validator(cls, elem) -> Optional[Validator]:
        raise NotImplementedError()

    @abstractmethod
    def _is_misconfigured(self):
        # NOTE method is called only when self.validator is not None and self.required is True
        raise NotImplementedError()


class Grid(ValidatedInput, ChoiceInput):
    """A Grid input element.

    Attributes:
        rows: The grid row names.
        cols: The grid columns (an alias for options).
        shuffle_rows: Shuffle this grid's rows.
    """

    _cell_symbols = ('?', '+')

    class _Index(ChoiceInput._Index):
        SHUFFLE_ROWS = 7
        VALIDATOR = 8
        ROW_NAME = 3
        MULTICHOICE = 11

    @staticmethod
    def parse_multichoice(elem):
        return bool(Grid._get_entries(elem)[0][Grid._Index.MULTICHOICE][0])

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        res.update({
            'rows': [entry[cls._Index.ROW_NAME][0] for entry in cls._get_entries(elem)],
            'shuffle_rows': bool(list_get(elem, cls._Index.SHUFFLE_ROWS, False)),
        })
        return res

    @classmethod
    def _parse_validator(cls, elem) -> Optional[GridValidator]:
        if len(elem) > cls._Index.VALIDATOR and elem[cls._Index.VALIDATOR]:
            return GridValidator.parse(elem[cls._Index.VALIDATOR][0])
        return None

    def __init__(self, *, rows, shuffle_rows, **kwargs):
        super().__init__(**kwargs)
        self.rows = rows
        self.cols = self.options  # alias
        self.shuffle_rows = shuffle_rows

    @property
    def options(self):
        return self._options[0]

    def _is_misconfigured(self):
        return self.validator.subtype is GridTypes.EXCLUSIVE_COLUMNS \
            and len(self.cols) < len(self.rows)

    def _hints(self, indent=0, modify=False):
        # NOTE row/column names may need wrapping
        res = super()._hints(indent, modify)
        max_length = max(len(row) for row in self.rows)
        header = ' ' * (max_length + 1) + '|'.join([opt.value for opt in self.options])
        row_fmt = f'{{:>{max_length}}} ' + \
                  ' '.join(f'{{:^{len(col.value)}}}' for col in self.cols)

        if not modify:
            cells = [[self._cell_symbols[0]] * len(self.options)] * len(self.rows)
        else:
            cells = [None] * len(self.rows)
            for i, row_choices in enumerate(self._values):
                cells[i] = [
                    self._cell_symbols[opt.value in row_choices]
                    for opt in self.options
                ]

        res.append(header)
        res += [row_fmt.format(row, *row_cells) for row, row_cells in zip(self.rows, cells)]
        return res

    def _answer(self):
        return []

    def _set_grid_values(self, values: Union[GridValue, EmptyValue]):
        if values is Value.EMPTY:
            return self._set_choices([Value.EMPTY] * len(self.rows))
        if not isinstance(values, list):
            raise ElementTypeError(self, values)
        if len(values) != len(self.rows):
            raise ElementValueError(self, values,
                                    details='Length of choices does not match the number of rows')
        choices = []
        for i, entry_value in enumerate(values):
            try:
                choices.append(self._to_choice_list(entry_value))
            except ElementTypeError as e:
                raise RowTypeError(self, e.value, index=i) from e
        try:
            self._set_choices(choices)
        except InvalidChoice as e:
            raise InvalidRowChoice(self, e.value, index=e.index) from e

    def _validate_entry(self, index):
        try:
            super()._validate_entry(index)
        except RequiredElement as e:
            raise RequiredRow(self, index=e.index)


class TimeInput(SingleInput):
    """An element which represents a time or a duration."""

    class _Index(SingleInput._Index):
        FLAGS = 6
        DURATION = 0

    @staticmethod
    def parse_duration_flag(elem):
        flags = TimeInput._get_entry(elem)[TimeInput._Index.FLAGS]
        return bool(flags[TimeInput._Index.DURATION])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._hour = None
        self._minute = None
        self._second = None

    def payload(self) -> Dict[str, List[str]]:
        """See base class."""
        payload = {}
        if self._hour is not None:
            payload[self._part_id('hour')] = self._hour
        if self._minute is not None:
            payload[self._part_id('minute')] = self._minute
        if self._second is not None:
            payload[self._part_id('second')] = self._second
        return payload

    def draft(self) -> List[Tuple]:
        """See base class."""
        hms = []
        for val in [self._hour, self._minute, self._second]:
            if val is not None:
                hms.append(f'{val:02}')
        # will return an empty string for an empty element.
        # The server sets the same draft value, so it's ok
        return [(None, self._entry_id, [':'.join(hms)], 0)]

    def _validate(self):
        if self.required and self._is_empty():
            raise RequiredElement(self)

    def _set_time(self, hour, minute, second):
        self._hour = hour
        self._minute = minute
        self._second = second

    def _is_empty(self):
        return all(val is None for val in [self._hour, self._minute, self._second])


class DateInput(SingleInput):
    """An element which represents a date.

    Attributes:
        has_year: Whether or not this element has a year field.
    """

    class _Index(SingleInput._Index):
        FLAGS = 7
        TIME = 0
        YEAR = 1

    @staticmethod
    def parse_time_flag(elem):
        flags = DateInput._get_entry(elem)[DateInput._Index.FLAGS]
        return bool(flags[DateInput._Index.TIME])

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        flags = cls._get_entry(elem)[cls._Index.FLAGS]
        res.update({
            'has_year': bool(flags[cls._Index.YEAR])
        })
        return res

    def __init__(self, *, has_year, **kwargs):
        super().__init__(**kwargs)
        self.has_year = has_year
        self._date = None

    def payload(self) -> Dict[str, List[str]]:
        """See base class."""
        if self._date is None:
            return {}
        payload = {
            self._part_id('month'): self._date.month,
            self._part_id('day'): self._date.day,
        }
        if self.has_year:
            payload[self._part_id("year")] = self._date.year
        return payload

    def draft(self) -> List[Tuple]:
        """See base class."""
        if self._date is None:
            return []
        fmt = '%Y-%m-%d' if self.has_year else '%m-%d'
        return [(None, self._entry_id, [self._date.strftime(fmt)], 0)]

    def _validate(self):
        if self.required and self._date is None:
            raise RequiredElement(self)

    def _answer(self) -> List[str]:
        if self._date is None:
            return ['EMPTY']
        fmt = '%m/%d'
        if self.has_year:
            fmt = '%Y/' + fmt
        return [f'"{self._date.strftime(fmt)}"']

    def _is_empty(self):
        return self._date is None


class MediaElement(Element):
    """A base class for media elements."""

    pass


class TextInput(ValidatedInput, SingleInput):
    """An element which accepts text input."""

    class _Index(SingleInput._Index):
        VALIDATOR = 4

    @classmethod
    def _parse_validator(cls, elem) -> Optional[TextValidator]:
        value = cls._get_entry(elem)
        if len(value) > cls._Index.VALIDATOR and value[cls._Index.VALIDATOR]:
            return TextValidator.parse(value[cls._Index.VALIDATOR][0])
        return None

    def set_value(self, value: Union[TextValue, EmptyValue]):
        """Sets the value for this element.

        Accepts a string or an empty value.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        if isinstance(value, str):
            if not value:
                value = Value.EMPTY
            else:
                return self._set_value([value])
        if value is Value.EMPTY:
            return self._set_value(value)
        raise ElementTypeError(self, value)

    def _is_misconfigured(self):
        return self.validator.subtype is NumberTypes.RANGE \
            and self.validator.args[0] > self.validator.args[1]


class ActionChoiceInput(ChoiceInput1D):
    """A ChoiceInput1D with optional actions based on choice.

    Attributes:
        next_page: The next page, based on user's choice.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.next_page = None

    def set_value(self, value: Union[ChoiceValue, EmptyValue]):
        """Sets the value for this element.

        Accepts an Option, a string or an empty value.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        return self._set_choices([self._to_choice_list(value)])

    def _set_choices(self, choices: List[Union[List[ChoiceValue], EmptyValue]]):
        self.next_page = None
        super()._set_choices(choices)
        # Don't reset no_loops if the element's options have no actions or the actions are ignored
        if (self._form is not None and
                isinstance(self.options[0], ActionOption) and
                cast(ActionOption, self.options[0]).next_page is not None):
            self._form._no_loops = False

    def _add_choice(self, choices, choice: Option):
        super()._add_choice(choices, choice)
        if isinstance(choice, ActionOption):
            self.next_page = choice.next_page

    def _answer(self) -> List[str]:
        from .elements import Page
        answer = super()._answer()
        if self.next_page is not None:
            if self.next_page is Page.SUBMIT:
                answer.append('Go to SUBMIT')
            else:
                answer.append(f'Go to page {self.next_page.index + 1}')
        return answer

    def _resolve_actions(self, next_page, mapping):
        for option in self.options:
            if isinstance(option, ActionOption):
                option._resolve_action(next_page, mapping)



import json
import re
from enum import Enum
from typing import Callable, Optional, Set, Dict, List
from urllib.parse import urlsplit, urlunsplit, parse_qs, urlencode

import requests
from requests.status_codes import codes
from bs4 import BeautifulSoup

from .elements_base import InputElement
from .elements import _Action, Element, Page, UserEmail, Value, parse as parse_element
from .elements import CallbackRetVal, default_callback
from .errors import ClosedForm, InfiniteLoop, ParseError, FormNotLoaded, FormNotValidated, \
    InvalidURL, NoSuchForm, EditingDisabled, SigninRequired
from .util import add_indent, page_separator, list_get


# Originally based on https://gist.github.com/gcampfield/cb56f05e71c60977ed9917de677f919c


CallbackType = Callable[[InputElement, int, int], CallbackRetVal]


class Settings:
    """Settings for a form.

    Attributes:
        collect_emails: Self-explanatory.
        send_receipt: Send a copy of user's responses to their e-mail address.
            If this value is SendReceipt.ALWAYS or the user opts in
            to receive the receipt, a captcha will be required
            to submit the form.
        signin_required: When Ture, the user must use a google account
            to submit this form. Only one response can be submitted.
        show_summary: Whether the user is allowed to view response stats.
        edit_responses: Self-explanatory.
        show_progressbar: Self-explanatory.
        shuffle_questions: Self-explanatory.
            It seems that the order of parsed questions isn't affected.
            However, shuffle_options on elements actually shuffles the options.
        show_resubmit_link: Self-explanatory.
        confirmation_msg:
            The message which is shown after a successful submission.
        is_quiz: Self-explanatory.
        immediate_grades: The user may view their grades (score) immediately.
        show_missed: "Identify which questions were answered incorrectly"
            (from the form creation page).
        show_correct_answers: Self-explanatory.
        show_points: Self-explanatory.
    """

    class _Index:
        FIRST_BLOCK = 2
        SECOND_BLOCK = 10
        QUIZ_BLOCK = 16

        # First block
        CONFIRMATION_MSG = 0
        SHOW_RESUBMIT_LINK = 1
        SHOW_SUMMARY = 2
        EDIT_RESPONSES = 3
        # Second block. The first 4 elements may be None
        SHOW_PROGRESSBAR = 0
        SIGNIN_REQUIRED = 1
        SHUFFLE_QUESTIONS = 2
        RECEIPT = 3
        COLLECT_EMAILS = 4
        # Quiz block
        # It's possible to create a form with IMMEDIATE_GRADES ==  COLLECT_EMAILS == 0
        GRADES_SETTINGS = 0
        SHOW_MISSED = 2
        SHOW_CORRECT = 3
        SHOW_POINTS = 4
        IMMEDIATE_GRADES = 1
        IS_QUIZ = 2

    class SendReceipt(Enum):
        UNUSED = None  # value is None when collect_emails is False
        OPT_IN = 1
        NEVER = 2
        ALWAYS = 3

    def parse(self, form_data):
        # any block may be missing
        first_block = form_data[self._Index.FIRST_BLOCK]
        second_block = form_data[self._Index.SECOND_BLOCK]
        quiz = list_get(form_data, self._Index.QUIZ_BLOCK, [])

        if first_block is not None:
            self.confirmation_msg = first_block[self._Index.CONFIRMATION_MSG] or None
            self.show_resubmit_link = bool(first_block[self._Index.SHOW_RESUBMIT_LINK])
            self.show_summary = bool(first_block[self._Index.SHOW_SUMMARY])
            self.edit_responses = bool(first_block[self._Index.EDIT_RESPONSES])
        if second_block is not None:
            self.show_progressbar = bool(second_block[self._Index.SHOW_PROGRESSBAR])
            self.signin_required = bool(second_block[self._Index.SIGNIN_REQUIRED])
            self.shuffle_questions = bool(second_block[self._Index.SHUFFLE_QUESTIONS])
            self.send_receipt = self.SendReceipt(second_block[self._Index.RECEIPT])
            self.collect_emails = bool(second_block[self._Index.COLLECT_EMAILS])
        self.is_quiz = bool(list_get(quiz, self._Index.IS_QUIZ, False))
        if self.is_quiz:
            self.immediate_grades = bool(quiz[self._Index.IMMEDIATE_GRADES])
            grades_settings = quiz[self._Index.GRADES_SETTINGS]
            self.show_missed = bool(grades_settings[self._Index.SHOW_MISSED])
            self.show_correct_answers = bool(grades_settings[self._Index.SHOW_CORRECT])
            self.show_points = bool(grades_settings[self._Index.SHOW_POINTS])

    def __init__(self):
        """Initializes all settings with a default value."""
        self.collect_emails = False
        self.send_receipt = self.SendReceipt.UNUSED

        self.signin_required = False

        self.show_summary = False
        self.edit_responses = False

        self.show_progressbar = False
        self.shuffle_questions = False
        self.show_resubmit_link = True

        self.confirmation_msg: Optional[str] = None

        self.is_quiz = False
        self.immediate_grades = True

        self.show_missed = True
        self.show_correct_answers = True
        self.show_points = True


class SubmissionResult:
    """A result of successful form submission.

    Attributes:
        resubmit: A link to submit another response.
        summary: A link to view summary charts.
        edit: A link to edit the response.
        quiz_score: A link to view the quiz score (if available).
    """

    def __init__(self, soup):
        self.resubmit = None
        self.summary = None
        self.edit = None
        self.quiz_score = None

        # Is the div class fixed?
        container = soup.find('div', class_='freebirdFormviewerViewResponseLinksContainer')
        for link in container.find_all('a'):
            href = link.get('href', '')
            if 'viewanalytics' in href:
                self.summary = href
            elif 'edit2' in href:
                self.edit = href
            elif 'viewscore' in href:
                self.quiz_score = href
            else:
                self.resubmit = href


class Form:
    """A Google Form wrapper.

    Attributes:
        url: URL that was used to load the form.
        name: Name of the form (== document name, not shown on the submission page).
        title: Title of the form.
        description: Description of the form.
        pages: List of form pages.
        settings: The form settings.
        is_loaded: Indicates if this form was properly loaded and parsed.
        is_validated:
            Indicates if this form was successfully validated and may be submitted.
    """

    class _DocIndex:
        FORM = 1
        NAME = 3
        URL = 14  # Unused. Is the index constant?
        SIGNIN_REQUIRED = 18  # Duplicate or has other meaning?

    class _FormIndex:
        DESCRIPTION = 0
        ELEMENTS = 1
        STYLE = 4  # Not implemented
        TITLE = 8

    url: Optional[str]
    name: Optional[str]
    title: Optional[str]
    description: Optional[str]
    pages: Optional[List[Page]]
    settings: Settings

    is_loaded: bool
    _unvalidated_elements: Set[Element]
    _no_loops: bool  # If true, the form is guaranteed to contain no loops.

    _prefilled_data: Dict[str, List[str]]
    _first_page: Optional[requests.models.Response]
    _fbzx: Optional[str]  # Doesn't need to be unique
    _history: Optional[str]
    _draft: Optional[str]

    @property
    def is_validated(self):
        """Indicates if all input elements were validated.

        If self.fill was called and did not raise an exception,
        this value will be True."""
        return len(self._unvalidated_elements) == 0 and self._no_loops

    def __init__(self):
        self._clear()

    def load(self, url, session: Optional[requests.Session] = None):
        """Loads and parses the form.

        Args:
            url: The form url. Usually looks like
                "https://docs.google.com/forms/.../viewform"
                or "https://forms.gle/...".
                Pre-filled links and response editing are also supported.
            session: A session which is used to load the form.
                If session is None, requests.get is used.

        Raises:
            gforms.errors.InvalidURL: The url is not a valid form url.
            gforms.errors.NoSuchForm: The form does not exist.
            requests.exceptions.RequestException: A request failed.
            gforms.errors.ParseError: The form could not be parsed.
            gforms.errors.ClosedForm: The form is closed.
            gforms.errors.EditingDisabled: Response editing is disabled.
            gforms.errors.SigninRequired:
                The form requires sign in to be loaded.
                (most probably, it contains a file upload element).
        """
        if session is None:
            session = requests

        self._clear()

        self.url = url

        self._first_page = session.get(self.url)
        self._check_resp(self._first_page)
        prefilled_data, is_edit = self._parse_url(self._first_page.url)
        self._prefilled_data = prefilled_data

        soup = BeautifulSoup(self._first_page.text, 'html.parser')

        data = self._raw_form(soup)
        self._fbzx = self._get_fbzx(soup)
        self._history = self._get_history(soup)
        self._draft = self._get_draft(soup)
        if data is None or any(value is None for value in [self._fbzx, self._history, self._draft]):
            raise ParseError(self)

        if is_edit:
            self._prefilled_data = self._prefill_from_draft(self._draft)

        self._parse(data)
        self.is_loaded = True

    def to_str(self, indent=0, include_answer=False):
        """Returns a text representation of the form.

        Args:
            indent: The indent for pages and elements.
            include_answer: A boolean,
                indicating if the output should contain the answers.

        Returns:
            A (multiline) string representation of this form.
        """
        if not self.is_loaded:
            raise FormNotLoaded(self)

        if self.description:
            title = f'{self.title}\n{self.description}'
        else:
            title = self.title
        separator = page_separator(indent)
        lines = '\n'.join(
            [separator] +
            [
                add_indent(page.to_str(indent=indent, include_answer=include_answer), indent) +
                '\n' + separator
                for page in self.pages
            ]
        )
        return f'{title}\n{lines}'

    def fill(
            self,
            callback: Optional[CallbackType] = None,
            fill_optional=False
    ):
        """Fills and validates the form using values returned by the callback.

        If callback is None, the default callback (see below) is used.

        For an example implementation of a callback,
        see gforms.elements.default_callback.

        For types and values accepted by an element, see its set_value method.

        You may also fill individual elements using their set_value method.
        In this ase, you will need to call Form.validate manually.

        Args:
            callback: The callback which returns values for elements.
                If the callback returns gforms.elements.Value.DEFAULT
                for an element, then the default callback
                is used for this element.

            fill_optional:
                Whether or not optional elements should be filled
                by the default callback.

        Raises:
            ValueError: The callback returned an invalid value.
            gforms.errors.ElementError:
                The callback returned an unexpected value for an element
            gforms.errors.ValidationError: The form cannot be submitted later
                if an element is filled with the callback return value
            gforms.errors.FormNotLoaded: The form was not loaded.
            gforms.errors.InfiniteLoop: The chosen values cannot be submitted,
                because the page transitions form an infinite loop.
            NotImplementedError: The default callback was used for an unsupported element.
        """
        self._iterate_elements(callback, fill_optional, do_fill=True)

    def validate(self):
        """Checks if the form can be submitted successfully.

        Raises:
            gforms.errors.FormNotLoaded: The form was not loaded.
            gforms.errors.ValidationError: An element has an invalid value.
            gforms.errors.InfiniteLoop:
                The page transitions form an infinite loop.
        """
        self._iterate_elements(do_fill=False)

    def submit(
            self,
            session=None, *,
            need_receipt=False,
            captcha_handler: Optional[Callable[[requests.models.Response], str]] = None,
            emulate_history=False
    ) -> SubmissionResult:
        """Submits the form.

        The form must be loaded, (optionally) filled and validated.

        Args:
            session: see Form.load
            need_receipt:
                Effective only if settings.send_receipt is SendReceipt.OPT_IN
                If True, will throw NotImplementedError.
            captcha_handler:
                A function which accepts a Response (a page with reCAPTCHA v2)
                and returns a string (g-recaptcha-response).
                NOTE Response and Response.request contain user's input.
                It should not be passed to third-party services unprocessed,
                since this may lead to privacy or security issues.
            emulate_history:
                Use only one request to submit the form
                (two requests for multipage forms with a captcha).
                For single-page forms the behavior is unchanged.
                For multi-page forms, the data from previous pages
                (pageHistory and draftResponse) is formed locally.

        Raises:
            requests.exceptions.RequestException: A request failed.
            RuntimeError: The predicted next page differs from the real one
                or a session(s) response with an incorrect code was received.
            ValueError: A captcha handler is required, but was not provided.
            gforms.errors.SigninRequired:
                The form requires sign in, this feature is not implemented.
            gforms.errors.ClosedForm: The form is closed.
            gforms.errors.EditingDisabled: Response editing is disabled.
            gforms.errors.NoSuchForm: The form does not exist (anymore).
            gforms.errors.FormNotLoaded: The form was not loaded.
            gforms.errors.FormNotValidated: The form was not validated.
        """
        if not self.is_loaded:
            raise FormNotLoaded(self)
        if not self.is_validated:
            raise FormNotValidated(self)

        if self.settings.send_receipt is not Settings.SendReceipt.OPT_IN:
            need_receipt = self.settings.send_receipt is Settings.SendReceipt.ALWAYS
        if need_receipt and captcha_handler is None:
            raise ValueError('captcha_handler is missing')

        if self.settings.signin_required:
            raise SigninRequired(self)

        if session is None:
            session = requests

        if emulate_history:
            last_response, page, history, draft = self._emulate_history(session, need_receipt)
        else:
            page = self.pages[0]
            history = self._history
            draft = self._draft
            last_response = self._first_page  # TODO prefill with sample values (at load time?)

        captcha_response = None

        while page is not None:
            next_page = page.next_page()
            # solve the CAPTCHA to submit the form and send a receipt
            if need_receipt and next_page is None:
                captcha_response = captcha_handler(last_response)
            last_response = self._submit_page(session, page, history, draft,
                                              continue_=next_page is not None,
                                              need_receipt=need_receipt,
                                              captcha_response=captcha_response)
            soup = BeautifulSoup(last_response.text, 'html.parser')
            history = self._get_history(soup)
            draft = self._get_draft(soup)
            if next_page is None and history is None:
                return SubmissionResult(soup)
            if next_page is None or history is None or \
                    next_page.index != int(history.rsplit(',', 1)[-1]):
                raise RuntimeError('Incorrect next page', self, last_response, next_page)
            page = next_page

    def _clear(self):
        self.url = None
        self.name = None
        self.title = None
        self.description = None
        self.pages = None
        self.settings = Settings()
        self.is_loaded = False

        self._prefilled_data = {}
        self._first_page = None

        self._unvalidated_elements = set()
        self._no_loops = True  # The form is guaranteed to contain no loops.

        self._fbzx = None  # Doesn't need to be unique
        self._history = None
        self._draft = None

    def _check_resp(self, resp):
        # will raise NoSuchForm for any 404/410 response. Check domain?
        if resp.status_code in [codes.not_found, codes.gone]:
            raise NoSuchForm(resp.url)
        if (
            resp.status_code == codes.unauthorized or  # submit, settings were updated
            urlsplit(resp.url).netloc == 'accounts.google.com'  # load, form contains a FileUpload
        ):
            raise SigninRequired(self)
        if self._is_closed(resp):
            if self.title is None:
                # Method is called from load(). Is the title needed?
                soup = BeautifulSoup(resp.text, 'html.parser')
                self.title = soup.find('title').text
            raise ClosedForm(self)
        if self._editing_disabled(resp):
            # if raised from Form.load, there is (?) no way to get the form title without additional requests
            raise EditingDisabled(self)

    @staticmethod
    def _parse_url(url: str):
        """Checks if the URL is a form url and extracts prefilled data."""
        url_data = urlsplit(url)
        if not url_data.path.endswith('viewform'):
            raise InvalidURL(url)
        prefilled_data = {}
        query = parse_qs(url_data.query)
        if query.get('usp', [''])[0] == 'pp_url':
            prefilled_data = {
                int(key[6:]): value
                for key, value in query.items() if key.startswith('entry.')
            }
        is_edit = 'edit2' in query
        return prefilled_data, is_edit

    @staticmethod
    def _response_url(url: str):
        url_data = list(urlsplit(url))
        url_data[2] = re.sub(r'viewform.*?$', 'formResponse', url_data[2])  # path
        original_query = parse_qs(url_data[3])
        query = {}
        # If 'edit2' is removed, it is possible to edit a response when editing is restricted.
        # This may be eventually fixed by Google.
        if 'edit2' in original_query:
            query['edit2'] = original_query['edit2']
        url_data[3] = urlencode(query, doseq=True)
        return urlunsplit(url_data)

    def _parse(self, data):
        self.name = data[self._DocIndex.NAME]
        form = data[self._DocIndex.FORM]
        self.title = form[self._FormIndex.TITLE]
        if not self.title:
            self.title = self.name
        self.description = form[self._FormIndex.DESCRIPTION]

        self.settings.parse(form)

        self.pages = [Page.first()]
        if self.settings.collect_emails:
            # Not a real element, but is displayed like one
            self._email_input = UserEmail()
            self._email_input.bind(self)
            self.pages[0].append(self._email_input)
            self._email_input.prefill(self._prefilled_data)

        if form[self._FormIndex.ELEMENTS] is None:
            return

        for elem in form[self._FormIndex.ELEMENTS]:
            el_type = Element.Type(elem[Element._Index.TYPE])
            if el_type == Element.Type.PAGE:
                self.pages.append(Page.parse(elem).with_index(len(self.pages)))
                # A multipage form with no input elements still needs to be validated.
                # A single-page form with action elements doesn't need additional validation
                # (actions are ignored).
                self._no_loops = False
                continue
            element = parse_element(elem)
            element.bind(self)
            self.pages[-1].append(element)
            if isinstance(element, InputElement):
                element.prefill(self._prefilled_data)
        self._resolve_actions()

    def _iterate_elements(
            self,
            callback: Optional[CallbackType] = None,
            fill_optional=False,
            *,
            do_fill
    ):
        """(optionally) fills and validates the elements."""
        if not self.is_loaded:
            raise FormNotLoaded(self)

        page = self.pages[0]
        pages_to_submit = {page}
        while page is not None:
            # use real element index or it would be better to count only input elements?
            for elem_index, elem in enumerate(page.elements):
                if not isinstance(elem, InputElement):
                    continue
                if do_fill:
                    value: CallbackRetVal = Value.DEFAULT
                    if callback is not None:
                        value = callback(elem, page.index, elem_index)
                        if value is None:  # missing return statement in the callback
                            raise ValueError('Callback returned an invalid value (None).'
                                             ' Is it missing a return statement?')
                    if value is Value.DEFAULT:
                        if elem.required or fill_optional:
                            value = default_callback(elem, page.index, elem_index)
                        else:
                            value = value.EMPTY
                    if value is not Value.UNCHANGED:
                        elem.set_value(value)

                if elem in self._unvalidated_elements:
                    elem.validate()

            page = page.next_page()
            if page in pages_to_submit:
                # It is possible to create a form in which any choice will lead to an infinite loop.
                # These cases are not detected (add a separate public method?)
                raise InfiniteLoop(self)
            pages_to_submit.add(page)
        self._no_loops = True

    def _resolve_actions(self):
        mapping = {page.id: page for page in self.pages}
        mapping[_Action.SUBMIT] = Page.SUBMIT
        for (page, next_page) in zip(self.pages, self.pages[1:] + [None]):
            page._resolve_actions(next_page, mapping)

    def _emulate_history(self, session, update_last_response):
        history = self._history.split(',')  # ['0']
        draft = json.loads(self._draft)  # basic draft, without prefill/edit: [None, None, fbzx]
        draft[0] = None  # may contain prefilled values
        # draft[0] should be None or non-empty,
        # but submission works with an empty value (on 25.05.21)

        # For some unfilled elements, draft values should be [""],
        # but if the corresponding entries are not included into the draft,
        # the form is still accepted

        if self.settings.collect_emails:
            if len(draft) < 8:
                draft += [None] * (8 - len(draft))
            draft[6] = self._email_input._value[0]
            draft[7] = 1  # ??

        last_page = self.pages[0]
        prev_page = None
        last_response = self._first_page

        while True:
            next_page = last_page.next_page()
            if next_page is None:
                if update_last_response and last_page is not self.pages[0]:
                    # TODO fill the form with sample values to get the last_response?
                    #   (see captcha_handler note in Form.submit)
                    last_response = self._submit_page(session,
                                                      prev_page,
                                                      ','.join(history[:-1]),
                                                      json.dumps(draft),
                                                      continue_=True, need_receipt=False)
                return last_response, last_page, ','.join(history), json.dumps(draft)
            history.append(str(next_page.index))
            if draft[0] is None:
                draft[0] = last_page.draft()
            else:
                draft[0] += last_page.draft()
            prev_page = last_page
            last_page = next_page

    def _submit_page(self, session, page, history, draft, *,
                     continue_, need_receipt, captcha_response=None):
        payload = page.payload()

        payload['fbzx'] = self._fbzx
        if continue_:
            payload['continue'] = 1
        payload['pageHistory'] = history
        payload['draftResponse'] = draft
        if need_receipt and not continue_:
            payload['g-recaptcha-response'] = captcha_response

        url = self._response_url(self._first_page.url)
        response = session.post(url, data=payload)
        self._check_resp(response)
        if response.status_code != 200:
            raise RuntimeError('Invalid response code', response)
        return response

    @staticmethod
    def _prefill_from_draft(draft: str):
        draft = json.loads(draft)
        if draft[0] is None:
            return {}
        prefilled_data = {}
        for entry in draft[0]:
            prefilled_data[entry[1]] = entry[2]  # Use an index class for draft?
        email = list_get(draft, 6, None)
        if email is not None:
            prefilled_data[UserEmail.ENTRY_ID] = [email]
        return prefilled_data

    @staticmethod
    def _is_closed(page):
        return page.url.endswith('closedform')

    @staticmethod
    def _editing_disabled(page):
        return page.url.endswith('editingdisabled')

    @staticmethod
    def _get_input(soup, name):
        elem = soup.find('input', {'name': name})
        if elem is None:
            return None
        return elem['value']

    @staticmethod
    def _get_fbzx(soup):
        return Form._get_input(soup, 'fbzx')

    @staticmethod
    def _get_history(soup):
        return Form._get_input(soup, 'pageHistory')

    @staticmethod
    def _get_draft(soup):
        return Form._get_input(soup, 'draftResponse')

    @staticmethod
    def _raw_form(soup):
        scripts = soup.find_all('script')
        pattern = re.compile(r'FB_PUBLIC_LOAD_DATA_ = (\[.+\])\n;', re.S)
        for script in scripts:
            if script.string is None:
                continue
            match = pattern.search(script.string)
            if match:
                return json.loads(match.group(1))
        return None



from __future__ import annotations

import random
from datetime import date, datetime, time, timedelta
from typing import List, Union, Optional, Dict, Tuple
from warnings import warn

from .elements_base import _Action, Value, MultiChoiceInput
from .elements_base import Element, MediaElement, InputElement, \
                           ActionChoiceInput, ChoiceInput1D, SingleInput, \
                           OtherChoiceInput, ValidatedInput, TextInput, \
                           DateInput, Grid, TimeInput
from .elements_base import ChoiceValue, EmptyValue, MultiChoiceValue, TextValue, \
                           GridChoiceValue, GridMultiChoiceValue

from .errors import ElementTypeError, InvalidDuration, \
                    InvalidText, MisconfiguredElement, \
                    UnknownValidator, InvalidArguments, UnknownElement
from .options import ActionOption
from .util import RADIO_SYMBOLS, CHECKBOX_SYMBOLS, add_indent, elem_separator, random_subset, \
    list_get
from .validators import CheckboxValidator, CheckboxTypes


__all__ = ['CallbackRetVal', 'Value',
           'Unknown', 'UserEmail', 'Page', 'Comment', 'Image', 'Video',
           'Short', 'Paragraph', 'Radio', 'Dropdown', 'Checkboxes', 'Scale',
           'RadioGrid', 'CheckboxGrid',
           'Time', 'Duration', 'Date', 'DateTime',
           'FileUpload',
           'default_callback', 'parse']


CheckboxesValue = MultiChoiceValue
ScaleValue = Union[ChoiceValue, int]
RadioGridValue = GridChoiceValue
CheckboxGridValue = GridMultiChoiceValue
DateValue = date
DateTimeValue = datetime
TimeValue = time
DurationValue = timedelta

ElemValue = Union[
    TextValue,  # Short, Paragraph
    ChoiceValue,  # Radio, Dropdown
    CheckboxesValue,
    ScaleValue,
    RadioGridValue,
    CheckboxGridValue,
    DateValue,
    DateTimeValue,
    TimeValue,
    DurationValue,
]

CallbackRetVal = Union[ElemValue, Value]


class Page(Element):
    """A Page element.

    Attributes:
        index: Index of the page.
        elements: A list of elements contained within the page.
    """

    class _Index(Element._Index):
        ACTION = 5

    SUBMIT: Optional[Page] = None

    @classmethod
    def first(cls):
        return cls(
            id_=_Action.FIRST,
            name=None,
            description=None,
            type_=Element.Type.PAGE,
            prev_action=_Action.NEXT,  # ignored
        ).with_index(0)

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        # FIRST / NEXT / SUBMIT / page id
        res['prev_action'] = list_get(elem, cls._Index.ACTION, _Action.NEXT)
        return res

    def __init__(self, *, prev_action, **kwargs):
        super().__init__(**kwargs)
        self._prev_action = prev_action
        self._next_page = None
        self._has_default_next_page = True
        self.index = None
        self.elements = []

    def with_index(self, index):
        self.index = index
        return self

    def append(self, elem):
        self.elements.append(elem)

    def next_page(self):
        """The next pge, based on user's choices."""
        if self._next_page is None:
            return None

        next_page = self._next_page
        for elem in self.elements:
            if isinstance(elem, ActionChoiceInput) and elem.next_page is not None:
                next_page = elem.next_page
        return next_page

    def to_str(self, indent=0, include_answer=False):
        """See base class."""
        title = f'Page {self.index + 1}:'
        if self.name:
            title = f'{title} {self.name}'
        if self.description:
            title = f'{title}\n{self.description}'
        if not self._has_default_next_page:
            if self._next_page is None:
                title = f'{title} -> Submit'
            else:
                title = f'{title} -> Page {self._next_page.index + 1}'
        if not self.elements:
            return title
        separator = elem_separator(indent)
        return '\n'.join(
            [title] +
            [
                separator + '\n' +
                add_indent(elem.to_str(indent=indent, include_answer=include_answer), indent)
                for elem in self.elements
            ]
        )

    def payload(self) -> Dict[str, List[str]]:
        """Returns a combined payload of all elements from this page."""
        payload = {}
        for elem in self.elements:
            if isinstance(elem, InputElement):
                payload.update(elem.payload())
        return payload

    def draft(self) -> List[Tuple]:
        """Returns a combined draftResponse."""
        result = []
        for elem in self.elements:
            if isinstance(elem, InputElement):
                result += elem.draft()
        return result

    def _resolve_actions(self, next_page: Optional[Page], mapping):
        for elem in self.elements:
            if isinstance(elem, ActionChoiceInput):
                elem._resolve_actions(next_page, mapping)

        if next_page is None:
            return
        if next_page._prev_action == _Action.SUBMIT:
            self._has_default_next_page = False
            return

        if next_page._prev_action == _Action.NEXT:
            self._next_page = next_page
        else:
            self._next_page = mapping[next_page._prev_action]
            self._has_default_next_page = False


Page.SUBMIT = Page(id_=_Action.SUBMIT, name=None, description=None,
                   type_=Element.Type.PAGE, prev_action=None).with_index(_Action.SUBMIT)


class Unknown(Element):
    pass


class Comment(Element):
    pass


class Image(MediaElement):
    # elem[6][0] - cosmoId (no direct link?)
    # search "google docs cosmoid" for more details
    # NOTE some input elements or options may contain attached images (parsing is not implemented)
    pass


class Video(MediaElement):
    class _Index(MediaElement._Index):
        VIDEO = 6
        LINK = 3

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        res.update({
            'link': elem[cls._Index.VIDEO][cls._Index.LINK]
        })
        return res

    def __init__(self, *, link, **kwargs):
        super().__init__(**kwargs)
        self._link = link

    def url(self):
        return f'https://youtu.be/{self._link}'

    def to_str(self, indent=0, **kwargs):
        s = super().to_str(indent, **kwargs)
        return f'{s}\n{self.url()}'


class Short(TextInput):
    """A single-line text input."""

    def _validate_entry(self, index):
        super()._validate_entry(index)
        if not self._value:
            return
        value = self._value[0]
        if '\n' in value:
            raise InvalidText(self, value.replace('\n', r'\n'), details='Input contains newlines')


class UserEmail(Short):
    """A virtual element used for e-mail collection."""

    ID = -9999
    ENTRY_ID = -999

    def __init__(self):
        from .validators import TextValidator, TextTypes
        val = TextValidator(type_=TextValidator.Type.TEXT, subtype=TextTypes.EMAIL,
                            error_msg='', args=None, bad_args=False)
        super().__init__(
            id_=self.ID,
            name='Email',
            description='Your email',
            type_=Element.Type.SHORT,
            entry_ids=[self.ENTRY_ID],
            required=True,
            validator=val
        )

    def _hints(self, indent=0, modify=False):
        # Don't include the validator hint
        from .elements_base import ValidatedInput
        return super(ValidatedInput, self)._hints(indent, modify)

    def payload(self):
        """See base class."""
        return {'emailAddress': self._value}

    def draft(self):
        """Returns empty list for Page.draft."""
        return []


class Paragraph(TextInput):
    """A multi-line text input."""

    pass


class Radio(ActionChoiceInput, OtherChoiceInput):
    _choice_symbols = RADIO_SYMBOLS

    def _resolve_actions(self, next_page, mapping):
        if isinstance(self.other_option, ActionOption):
            self.other_option._resolve_action(next_page, mapping)
        super()._resolve_actions(next_page, mapping)


class Dropdown(ActionChoiceInput):
    pass


class Checkboxes(ValidatedInput, OtherChoiceInput, MultiChoiceInput):
    _choice_symbols = CHECKBOX_SYMBOLS

    class _Index(ChoiceInput1D._Index):
        VALIDATOR = 4

    @classmethod
    def _parse_validator(cls, elem) -> Optional[CheckboxValidator]:
        value = cls._get_entry(elem)
        if len(value) > cls._Index.VALIDATOR and value[cls._Index.VALIDATOR]:
            return CheckboxValidator.parse(value[cls._Index.VALIDATOR][0])
        return None

    def set_value(self, value: Union[CheckboxesValue, EmptyValue]):
        """Sets the value for this element.

        Accepts a choice value (a string or an Option),
        a list of choice values or an empty value.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        return self._set_choices([self._to_choice_list(value)])

    def _is_misconfigured(self):
        cnt = len(self.options)
        if self.other_option is not None:
            cnt += 1
        required_cnt = self.validator.args[0]  # always > 0
        if (self.validator.subtype is CheckboxTypes.AT_LEAST
                or self.validator.subtype is CheckboxTypes.EXACTLY) and cnt < required_cnt:
            return True
        return False


class Scale(ChoiceInput1D):
    """A liner scale.

    Attributes:
        low: The label for the lowest value (may be empty).
        high: The label for the highest value (may be empty).
    """

    class _Index(ChoiceInput1D._Index):
        LABELS = 3

    @classmethod
    def _parse(cls, elem):
        res = super()._parse(elem)
        low, high = cls._get_entry(elem)[cls._Index.LABELS]
        res.update({
            'low': low,
            'high': high,
        })
        return res

    def __init__(self, *, low, high, **kwargs):
        super().__init__(**kwargs)
        self.low = low
        self.high = high

    def set_value(self, value: Union[ScaleValue, EmptyValue]):
        """Sets the value for this element.

        Accepts a string, an Option, an int or an empty value.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        if isinstance(value, int):
            value = str(value)
        return self._set_choices([self._to_choice_list(value)])

    def _hints(self, indent=0, modify=False):
        max_len = 1 if len(self.options) < 10 else 2
        tab = ' ' * (len(self.low) + 1) if self.low else ''
        header = tab + ' '.join([f'{opt.value:^{max_len}}' for opt in self.options])
        scale = [f'{RADIO_SYMBOLS[0]:^{max_len}}' for _ in self.options]
        if modify and self._value:
            scale[int(self._value[0]) - 1] = f'{RADIO_SYMBOLS[1]:^{max_len}}'
        scale = ' '.join(scale)
        if self.low:  # wrapping?
            scale = f'{self.low} {scale}'
        if self.high:
            scale = f'{scale} {self.high}'
        return [header, scale]

    def _answer(self):
        return []


class RadioGrid(Grid):
    _cell_symbols = RADIO_SYMBOLS

    def set_value(self, value: Union[RadioGridValue, EmptyValue]):
        """Sets the value for this element.

        Accepts an empty value or a list of choice values,
        each of which is a string, an Option or an empty value

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
            gforms.errors.ElementValueError:
                The argument is a list,
                but its length does not match the number of rows.
        """
        self._set_grid_values(value)


class CheckboxGrid(Grid, MultiChoiceInput):
    _cell_symbols = CHECKBOX_SYMBOLS

    def set_value(self, value: Union[CheckboxGridValue, EmptyValue]):
        """Sets the value for this element.

        Accepts the same values as RadioGrid, but also
        allows each list element to be a list of strings or Option objects.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
            gforms.errors.ElementValueError:
                The argument is a list,
                but its length does not match the number of rows.
        """
        self._set_grid_values(value)


class Time(TimeInput):
    def set_value(self, value: Union[TimeValue, EmptyValue]):
        """Sets the value for this element.

        Accepts a datetime.time or an empty value.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        if value is Value.EMPTY:
            return self._set_time(None, None, None)
        if isinstance(value, time):
            return self._set_time(value.hour, value.minute, None)
        raise ElementTypeError(self, value)

    def _answer(self) -> List[str]:
        if self._is_empty():
            return ['EMPTY']
        return [f'{self._hour:02}:{self._minute:02}']


class Duration(TimeInput):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._timedelta = None

    def set_value(self, value: Union[DurationValue, EmptyValue]):
        """Sets the value for this element.

        Accepts a datetime.timedelta or an empty value.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        if value is Value.EMPTY:
            self._set_time(None, None, None)
            self._timedelta = None
        elif isinstance(value, timedelta):
            seconds = int(value.total_seconds())
            h, m, s = seconds // 3600, seconds // 60 % 60, seconds % 60
            self._set_time(h, m, s)
            self._timedelta = value
        else:
            raise ElementTypeError(self, value)

    def _validate(self):
        super()._validate()
        if self._timedelta is None:
            return
        seconds = self._timedelta.total_seconds()
        if seconds >= 73 * 3600 or seconds < 0:
            raise InvalidDuration(self, self._timedelta)

    def _answer(self) -> List[str]:
        if self._is_empty():
            return ['EMPTY']
        h = self._hour or 0
        m = self._minute or 0
        s = self._second or 0
        return [f'{h:02}:{m:02}:{s:02}']


class Date(DateInput):
    def set_value(self, value: Union[DateValue, EmptyValue]):
        """Sets the value for this element.

        Accepts a datetime.date or an empty value.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        if value is Value.EMPTY:
            self._date = None
        elif isinstance(value, date):
            self._date = value
        else:
            raise ElementTypeError(self, value)


class DateTime(DateInput):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._time = None

    def set_value(self, value: Union[DateTimeValue, EmptyValue]):
        """Sets the value for this element.

        Accepts a datetime.datetime or an empty value.

        Args:
            value: The value for this element.

        Raises:
            gforms.errors.ElementTypeError:
                The argument's type is not accepted by this element.
        """
        if value is Value.EMPTY:
            date_ = time_ = None
        elif isinstance(value, datetime):
            date_, time_ = value.date(), value.time()
        else:
            raise ElementTypeError(self, value)
        # self._date is None if and only if self._time is None, no need to extend validation
        self._date = date_
        self._time = time_

    def payload(self) -> Dict[str, List[str]]:
        """See base class."""
        if self._date is None:
            return {}
        payload = super().payload()
        payload.update({
            self._part_id("hour"): self._time.hour,
            self._part_id("minute"): self._time.minute,
        })
        return payload

    def draft(self) -> List[Tuple]:
        """See base class."""
        if self._date is None:
            return []
        result = super().draft()
        date_str = result[0][2][0]  # add DraftIndex class?
        result[0][2][0] = date_str + f' {self._time.strftime("%H:%M")}'
        return result

    def _answer(self) -> List[str]:
        if self._date is None:
            return ['EMPTY']
        answer = super()._answer()
        answer[0] += f' {self._time.strftime("%H:%M")}'
        return answer


class FileUpload(SingleInput):
    # As of June 2021, forms with a file upload require sign in for loading,
    # so this class is actually never used
    def set_value(self, value):
        raise NotImplementedError()


def _grid_validated_choices(elem: Grid):
    n = len(elem.rows)
    if isinstance(elem, RadioGrid):
        if elem.required:
            # cols >= rows (checked in default_callback by elem.is_misconfigured)
            # Just pick a unique option for each row.
            return random.sample(elem.options, k=n)
        # Pick some options and put each into a unique random row.
        # With prob=0.5 the grid is too empty.
        opts = random_subset(elem.options, max_size=n, prob=0.8)
        row_indices = random.sample(range(n), k=len(opts))
        choices = [Value.EMPTY for _ in range(n)]
        for opt, idx in zip(opts, row_indices):
            choices[idx] = opt
        return choices

    # CheckboxGrid
    result = [[] for _ in range(n)]
    opts = random_subset(elem.options,
                         min_size=n if elem.required else 0,
                         prob=0.8)
    if elem.required:
        result = [[opt] for opt in opts[:n]]  # choose at least one option in each row
        opts = opts[n:]

    # Add each of the remaining options to a random row
    row_indices = random.choices(range(n), k=len(opts))
    for opt, idx in zip(opts, row_indices):
        result[idx].append(opt)
    return result


def _cb_validated_choices(elem: Checkboxes):
    required_cnt = elem.validator.args[0]
    if elem.validator.subtype is CheckboxTypes.AT_MOST:
        return random_subset(elem.options, min_size=int(elem.required), max_size=required_cnt)
    if len(elem.options) + int(elem.other_option is not None) < required_cnt:
        # Impossible to select the required number of options.
        # Element is optional, this is the only valid value
        return Value.EMPTY
    if len(elem.options) < required_cnt:
        if not elem.required:
            return Value.EMPTY
        raise NotImplementedError('Need to choose "Other" value to meet validation requirements')
    if elem.validator.subtype is CheckboxTypes.AT_LEAST:
        return random_subset(elem.options, min_size=required_cnt)
    if elem.validator.subtype is CheckboxTypes.EXACTLY:
        return random.sample(elem.options, required_cnt)


def _validated_choices(elem):
    if isinstance(elem, Grid):
        return _grid_validated_choices(elem)
    if isinstance(elem, Checkboxes):
        return _cb_validated_choices(elem)
    raise NotImplementedError()


def default_callback(elem: InputElement, page_index, elem_index) -> Union[ElemValue, EmptyValue]:
    """The default callback implementation for Form.fill.

    This callback will raise a NotImplementedError
    if it is called on a TextInput, DateInput or a TimeInput.call
    """

    # Single choice inputs
    if isinstance(elem, Scale) or isinstance(elem, Dropdown) or isinstance(elem, Radio):
        opts = elem.options[:]
        if not elem.required:
            opts.append(Value.EMPTY)
        return random.choice(opts)

    # Example: allow "Other" in Radio
    #     if elem.other_option is not None:
    #         opt = random.choice(elem.options + [elem.other_option])
    #         if opt.other:
    #             # return 'Sample text'  # another alternative
    #             opt.value = 'Sample text'
    #         return opt

    # Check the validator; for TextInput raise NotImplementedError (later)
    if isinstance(elem, ValidatedInput) and not isinstance(elem, TextInput) \
            and elem.validator is not None:

        if elem.is_misconfigured():
            raise MisconfiguredElement(elem)
        if elem.validator.has_unknown_type():
            # Raise, not warn. Otherwise, most probably, form submission will fail
            # If needed, custom callback should handle such cases
            raise UnknownValidator(
                type(elem.validator), [elem.validator.type, elem.validator.subtype]
            )
        if elem.validator.bad_args:
            raise InvalidArguments(
                type(elem.validator), elem.validator.type,
                elem.validator.subtype, elem.validator.args
            )
        return _validated_choices(elem)

    if isinstance(elem, Checkboxes):
        # Don't auto-choose "Other"
        return random_subset(elem.options, min_size=int(elem.required))

    if isinstance(elem, Grid):
        n = len(elem.rows)
        if isinstance(elem, CheckboxGrid):
            return [random_subset(elem.options, min_size=int(elem.required)) for _ in range(n)]
        opts = elem.options
        if not elem.required:
            opts = opts + [Value.EMPTY]
        return [random.choice(opts) for _ in range(n)]

    # text inputs / Date / DateTime / Time / Duration
    raise NotImplementedError(f'Cannot choose a random value for {elem._type_str()}')


_element_mapping = {getattr(Element.Type, el_type.__name__.upper()): el_type for el_type in [
    Unknown,
    Short, Paragraph,
    Radio, Dropdown, Checkboxes, Scale,
    Comment, Page, Image, Video,
]}

_element_mapping[Element.Type.FILE_UPLOAD] = FileUpload


def parse(elem):
    """Creates an Element of the right type from its JSON representation."""
    el_type = Element.Type(elem[Element._Index.TYPE])
    if el_type is Element.Type.GRID:
        cls = CheckboxGrid if Grid.parse_multichoice(elem) else RadioGrid
    elif el_type is Element.Type.DATE:
        cls = DateTime if DateInput.parse_time_flag(elem) else Date
    elif el_type is Element.Type.TIME:
        cls = Duration if TimeInput.parse_duration_flag(elem) else Time
    else:
        cls = _element_mapping[el_type]
    if el_type is Element.Type.UNKNOWN:
        warn(UnknownElement(elem, elem[Element._Index.TYPE]))
    return cls.parse(elem)



from abc import ABC, abstractmethod


class FormsError(Exception, ABC):
    """Base error class"""

    def __init__(self, *, details=None):
        super().__init__()
        self.details = details

    @abstractmethod
    def _message(self):
        raise NotImplementedError()

    def __str__(self):
        message = self._message()
        if self.details:
            message += f' ({self.details})'
        return message


class InvalidURL(FormsError):
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url

    def _message(self):
        return f'Not a valid form URL: "{self.url}".'


class NoSuchForm(InvalidURL):
    def _message(self):
        return f'A form with URL "{self.url}" does not exist.'


class FormError(FormsError):
    def __init__(self, form, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form = form

    @property
    def form_descr(self):
        if self.form.title is not None:
            return f'"{self.form.title}"'
        elif self.form.url is not None:
            return 'with URL ' + self.form.url
        return ''


class ParseError(FormError):
    def _message(self):
        return 'Cannot parse form ' + self.form_descr


class ClosedForm(ParseError):
    def _message(self):
        return f'Form {self.form_descr} is closed'


class SigninRequired(FormError, NotImplementedError):
    def _message(self):
        return f'Form {self.form_descr} requires sign in'


class EditingDisabled(ParseError):
    def _message(self):
        return 'Cannot edit a response for form ' + self.form_descr


class FormNotLoaded(FormError):
    def _message(self):
        return f'Form {self.form_descr} was not loaded'


class FormNotValidated(FormError):
    def _message(self):
        return f'The form needs to be validated.'


class ElementError(FormsError):
    def __init__(self, elem, *args, index=0, **kwargs):
        super().__init__(*args, **kwargs)
        self.elem = elem
        self.index = index


class InfiniteLoop(FormError, ValueError):
    def __init__(self, form):
        super().__init__(form)

    def _message(self):
        return 'Chosen input values lead to an infinite loop'


class RowError(ElementError):
    @property
    def row(self):
        return self.elem.rows[self.index]


class ValidationError(ElementError, ValueError):
    pass


class InvalidValue(ValidationError):
    def __init__(self, elem, value, *args, **kwargs):
        super().__init__(elem, *args, **kwargs)
        self.value = value


class MisconfiguredElement(ValidationError):
    def _message(self):
        return f'The validator requirements for element "{self.elem.name}" are:' \
               f' "{self.elem.validator.to_str()}". A valid value for this element does not exist.'


# The following errors may be raised from InputElement.set_value

class ElementTypeError(ElementError, TypeError):
    def __init__(self, elem, value, *args, **kwargs):
        super().__init__(elem, *args, **kwargs)
        self.value = value

    def _message(self):
        return 'Unsupported argument type' \
               f' (element: "{self.elem.name}", argument: {repr(self.value)})'


class ElementValueError(ElementError, ValueError):  # not the best name
    def __init__(self, elem, value, *args, **kwargs):
        super().__init__(elem, *args, **kwargs)
        self.value = value

    def _message(self):
        return 'Cannot fill an entry with the chosen value' \
               f' (element: "{self.elem.name}", value: {repr(self.value)})'


class InvalidChoice(ElementValueError):
    def _message(self):
        return f'Invalid choice in "{self.elem.name}" ({repr(self.value)})'


class DuplicateOther(ElementValueError):
    def __init__(self, elem, value, value2, *args, **kwargs):
        super().__init__(elem, value, *args, **kwargs)
        self.value2 = value2

    def _message(self):
        return f'Duplicate "Other" values in "{self.elem.name}"' \
               f' ("{self.value}" and "{self.value2}")'


class RowTypeError(ElementTypeError, RowError):
    def _message(self):
        return 'Unsupported argument type' \
               f' (element: "{self.elem.name}", row {self.row}, argument: {repr(self.value)})'


class InvalidRowChoice(InvalidChoice, RowError):
    def _message(self):
        return f'Invalid choice in "{self.elem.name}" ({self.value}) (row {self.row})'


# The following errors may be raised from InputElement.validate

class RequiredElement(ValidationError):
    def _message(self):
        return f'An entry in required element "{self.elem.name}" is empty'


class EmptyOther(RequiredElement):
    def _message(self):
        return f'Empty "Other" value in required element "{self.elem.name}"'


class RequiredRow(RequiredElement, RowError):
    def _message(self):
        return f'A row ("{self.row}") in required element "{self.elem.name}" is empty'


class InvalidText(InvalidValue):
    def _message(self):
        return f'Invalid text input in "{self.elem.name}" ("{self.value}")'


class InvalidDuration(InvalidValue):
    def _message(self):
        return f'Duration value ({self.value.total_seconds()} seconds) is out of range:' \
               ' value should be positive and less than 73 hours'


class UnknownElement(UserWarning):
    def __init__(self, el_data, el_type):
        super().__init__()
        self.el_data = el_data
        self.el_type = el_type

    def __str__(self):
        return f'Found an element with unknown type ({self.el_type}). Raw element data: {self.el_data}'


class ValidatorWarning(UserWarning):
    pass


class UnknownValidator(ValidatorWarning):
    def __init__(self, cls, value):
        super().__init__()
        self.cls = cls
        self.value = value

    def __str__(self):
        return f'Unknown validator type for {self.cls.__name__}: {repr(self.value)},' \
               ' this validator will allow any input'


class InvalidArguments(ValidatorWarning):
    def __init__(self, cls, val_type, subtype, args):
        super().__init__()
        self.cls = cls
        self.val_type = val_type
        self.subtype = subtype
        self.args = args

    def __str__(self):
        return f'Cannot convert arguments for {self.cls.__name__}.' \
               f' Type: {self.val_type} ({self.subtype}), Arguments: {repr(self.args)}.' \
               ' Validator is disabled'


class SameColumn(ValidationError):
    def __init__(self, elem, column=None, *args, **kwargs):
        super().__init__(elem, *args, **kwargs)
        self.column = column

    def _message(self):
        detail = f'("{self.column}") ' if self.column else ''
        return f'Same column {detail}is chosen in two or more rows of "{self.elem}".' \
               ' The choices must be unique'


class InvalidChoiceCount(ValidationError):
    def __init__(self, elem, cnt, *args, **kwargs):
        super().__init__(elem, *args, **kwargs)
        self.cnt = cnt

    def _message(self):
        val = self.elem.validator
        return f'{self.cnt} option(s) were chosen in "{self.elem.name}",' \
               f' but {val.subtype.descr(val.args)} are required'



import re
from setuptools import setup, find_packages


with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()


with open('gforms/__init__.py', encoding='utf-8') as f:
    version = re.search(r"__version__ = '(.+)'", f.read()).group(1)


setup(
    name='gforms',
    description='Google Forms wrapper for Python',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='vvd170501',
    url='https://github.com/vvd170501/python-gforms',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    packages=['gforms'],
    version=version,
    license_files=('LICENSE',),

    python_requires='>=3.6',
    install_requires=[
        'beautifulsoup4',
        'requests',
    ],
    extras_require={
        'dev': [
            'pytest',
        ]
    },
)



import re
from abc import ABC, abstractmethod
from collections import Counter
from typing import TYPE_CHECKING, List, cast
from warnings import warn

from .errors import SameColumn, UnknownValidator, InvalidText, InvalidArguments, InvalidChoiceCount
from .util import EMAIL_REGEX, URL_REGEX, DefaultEnum, ArgEnum, list_get

if TYPE_CHECKING:
    from elements import Checkboxes


class Subtype(DefaultEnum, ArgEnum):
    # Add UNKNOWN dynamically?

    @property
    def argnum(self):
        return self.arg[0]

    def descr(self, args):
        if args is None:
            return self.arg[1]
        return self.arg[1].format(*args)


class Validator(ABC):
    """A validator for a form element.

    Attributes:
        type: A member of cls.Type, indicating the validator's type.
        subtype: The validator's subtype.
        args: Validator arguments.
        error_msg: Custom error message, if it exists.
        bad_args: A boolean
            indicating if the arguments could not be parsed correctly.
    """

    class Type(DefaultEnum, ArgEnum):
        # Add UNKNOWN dynamically?

        @property
        def subtype_class(self):
            return self.arg

    class _Index:
        TYPE = 0
        SUBTYPE = 1
        ARGS_OR_MSG = 2

    @classmethod
    def parse(cls, val):
        """Creates a Validator from its JSON representation."""
        if len(val) <= cls._Index.SUBTYPE:
            return cls._unknown_validator(val)
        type_ = cls.Type(val[cls._Index.TYPE])
        if type_ is cls.Type.UNKNOWN:
            return cls._unknown_validator(val)
        subtype_cls = type_.subtype_class
        subtype = subtype_cls(val[cls._Index.SUBTYPE])
        if subtype is subtype_cls.UNKNOWN:
            return cls._unknown_validator(val, type_, subtype)

        argnum = subtype.argnum
        msg_index = cls._Index.ARGS_OR_MSG
        args = None
        if argnum > 0:
            args = list_get(val, cls._Index.ARGS_OR_MSG, None)
            msg_index += 1
        # If args are needed but not found, msg_index will not be increased. msg is not used -> ok
        msg = list_get(val, msg_index, None)
        args, valid_args = cls._parse_args(args, type_, subtype, argnum)
        if not valid_args:
            warn(InvalidArguments(cls, type_, subtype, args))
        return cls(type_, subtype, args, not valid_args, msg)

    def __init__(self, type_, subtype, args, bad_args, error_msg):
        self.type = type_
        self.subtype: Subtype = subtype
        self.args = args
        self.error_msg = error_msg
        self.bad_args = bad_args

    def validate(self, elem):
        """Validates the element's value."""
        if self.has_unknown_type() or self.bad_args:
            return
        self._validate(elem, elem._values)

    def has_unknown_type(self):
        return self.type is self.Type.UNKNOWN or self.subtype is self.type.subtype_class.UNKNOWN

    def to_str(self):
        if self.has_unknown_type():
            return 'Unknown validator'
        return f'{self._descr()}'

    @classmethod
    def _unknown_validator(cls, val, type_=None, subtype=None):
        warn(UnknownValidator(cls, val))
        return cls(type_ or cls.Type.UNKNOWN, subtype or cls.Type.UNKNOWN,
                   args=None, bad_args=False, error_msg=None)  # only type is important

    @classmethod
    def _parse_args(cls, args, type_, subtype, argnum):
        """
        Convert arguments to needed type. All args are received as strings
        Return value: converted args and a value indicating if conversion was successful
        """
        if args is None:
            return None, not argnum  # Error if args are required, but not found
        if not isinstance(args, list):
            return args, False  # e.g. new argnum is zero -> we get err_msg instead of args
        if len(args) != argnum:
            return args, False
        return cls._parse_arg_list(args, type_, subtype)

    @classmethod
    @abstractmethod
    def _parse_arg_list(cls, args, type_, subtype):
        raise NotImplementedError()

    def _descr(self):
        return self.subtype.descr(self.args)

    @abstractmethod
    def _validate(self, elem, values: List[List[str]]):
        raise NotImplementedError()


class NumberTypes(Subtype):
    UNKNOWN = (-1, (0, 'Unknown validator'))
    GT = (1, (1, 'Number > {}'))
    GE = (2, (1, 'Number >= {}'))
    LT = (3, (1, 'Number < {}'))
    LE = (4, (1, 'Number <= {}'))
    EQ = (5, (1, 'Number == {}'))
    NE = (6, (1, 'Number != {}'))
    RANGE = (7, (2, 'Number in range [{}, {}]'))
    NOT_RANGE = (8, (2, 'Number not in range [{}, {}]'))
    IS_NUMBER = (9, (0, 'A number'))
    IS_INT = (10, (0, 'An integer'))


class TextTypes(Subtype):
    UNKNOWN = (-1, (0, 'Unknown validator'))
    CONTAINS = (100, (1, 'Must contain "{}"'))
    NOT_CONTAINS = (101, (1, 'Must not contain "{}"'))
    EMAIL = (102, (0, 'An E-mail address'))
    URL = (103, (0, 'URL'))


class LengthTypes(Subtype):
    UNKNOWN = (-1, (0, 'Unknown validator'))
    MAX_LENGTH = (202, (1, 'Max {} characters'))
    MIN_LENGTH = (203, (1, 'Min {} characters'))


class RegexTypes(Subtype):
    UNKNOWN = (-1, (0, 'Unknown validator'))
    CONTAINS = (299, (1, 'Must contain regex "{}"'))
    NOT_CONTAINS = (300, (1, 'Must not contain regex "{}"'))
    MATCHES = (301, (1, 'Must match regex "{}"'))
    NOT_MATCHES = (302, (1, 'Must not match regex "{}"'))

    def descr(self, args):
        if args is None:
            return self.arg[1]
        return self.arg[1].format(*(arg.pattern for arg in args))


class TextValidator(Validator):
    class Type(Validator.Type):
        UNKNOWN = (-1, None)
        NUMBER = (1, NumberTypes)
        TEXT = (2, TextTypes)
        REGEX = (4, RegexTypes)
        LENGTH = (6, LengthTypes)

    @classmethod
    def _parse_arg_list(cls, args, type_, subtype):
        if type_ is cls.Type.TEXT:
            return args, True  # only string args
        if type_ is cls.Type.LENGTH:
            try:
                return [int(arg) for arg in args], True
            except ValueError:
                return args, False
        if type_ is cls.Type.NUMBER:
            try:
                args = [float(arg) for arg in args]
                for i in range(len(args)):
                    if args[i].is_integer():
                        args[i] = int(args[i])
                return args, True
            except ValueError:
                return args, False
        if type_ is cls.Type.REGEX:
            try:
                return [re.compile(arg) for arg in args], True
            except re.error:
                return args, False

    def _descr(self):
        return 'Allowed values: ' + super()._descr()

    def _validate(self, elem, values: List[List[str]]):
        if not values[0]:  # empty element
            return
        value = values[0][0]
        if self.type is self.Type.NUMBER:
            is_ok = self._validate_number(value)
        elif self.type is self.Type.TEXT:
            is_ok = self._validate_text(value)
        elif self.type is self.Type.LENGTH:
            is_ok = self._validate_length(value)
        elif self.type is self.Type.REGEX:
            is_ok = self._validate_regex(value)
        else:
            raise NotImplementedError()
        if not is_ok:
            descr = self._descr()
            if self.error_msg:
                descr = f'{self.error_msg} ({descr})'
            raise InvalidText(elem, value, details=descr)

    def _validate_number(self, val):
        if self.subtype is NumberTypes.IS_INT:
            try:
                _ = int(val)
                return True
            except ValueError:
                return False
        try:
            val = float(val)
        except ValueError:
            return False

        if self.subtype is NumberTypes.IS_NUMBER:
            # NOTE will return True for "NaN", "Inf" and "Infinity".
            # JS validator accepts only the last one.
            # It was not tested if these values pass the server-side validation
            return True
        arg = self.args[0]
        if self.subtype is NumberTypes.GT:
            return val > arg
        if self.subtype is NumberTypes.GE:
            return val >= arg
        if self.subtype is NumberTypes.LT:
            return val < arg
        if self.subtype is NumberTypes.LE:
            return val <= arg
        if self.subtype is NumberTypes.EQ:
            # if arg == 2.0, "1.99..9" will pass validation.
            # It's ok, as long as the value is accepted by server
            return val == arg
        if self.subtype is NumberTypes.NE:
            return val != arg
        if self.subtype is NumberTypes.RANGE:
            # The first arg can be greater than the second. This is checked in TextInput.
            return self.args[0] <= val <= self.args[1]
        if self.subtype is NumberTypes.NOT_RANGE:
            # The first arg can be greater than the second, it's ok
            return val < min(self.args) or val > max(self.args)
        raise NotImplementedError()

    def _validate_text(self, val):
        if self.subtype is TextTypes.CONTAINS:
            return self.args[0] in val
        if self.subtype is TextTypes.NOT_CONTAINS:
            return self.args[0] not in val
        if self.subtype is TextTypes.EMAIL:
            return EMAIL_REGEX.match(val) is not None
        if self.subtype is TextTypes.URL:
            # couldn't find a regex / checking algorithm in page source
            # It seems that URLS are validaetd on the server side
            # (client performs only basic checks)
            return URL_REGEX.match(val) is not None
        raise NotImplementedError()

    def _validate_length(self, val):
        if self.subtype is LengthTypes.MIN_LENGTH:
            return len(val) >= self.args[0]
        if self.subtype is LengthTypes.MAX_LENGTH:
            return len(val) <= self.args[0]
        raise NotImplementedError()

    def _validate_regex(self, val):
        if self.subtype is RegexTypes.CONTAINS:
            return self.args[0].search(val) is not None
        if self.subtype is RegexTypes.NOT_CONTAINS:
            return self.args[0].search(val) is None
        if self.subtype is RegexTypes.MATCHES:
            return self.args[0].match(val) is not None
        if self.subtype is RegexTypes.NOT_MATCHES:
            return self.args[0].match(val) is None
        raise NotImplementedError()


class GridTypes(Subtype):
    UNKNOWN = (-1, (0, 'Unknown validator'))
    EXCLUSIVE_COLUMNS = (205, (0, 'Max 1 response per column'))


class GridValidator(Validator):
    class Type(Validator.Type):
        UNKNOWN = (-1, None)
        DEFAULT = (8, GridTypes)

    @classmethod
    def _parse_arg_list(cls, args, type_, subtype):
        # no args are needed.
        # This method will be called only if args is an empty list (is it possible?)
        return None, True

    def _validate(self, elem, values: List[List[str]]):
        if self.subtype is GridTypes.EXCLUSIVE_COLUMNS:
            cnt = Counter()
            for row in values:
                if row:
                    cnt.update(row)
                    col, count = max(cnt.items(), key=lambda x: x[1])
                    if count > 1:
                        raise SameColumn(elem, col)
        else:
            raise NotImplementedError()


class CheckboxTypes(Subtype):
    UNKNOWN = (-1, (0, 'Unknown validator'))
    AT_LEAST = (200, (1, 'at least {} option(s)'))
    AT_MOST = (201, (1, 'at most {} option(s)'))
    EXACTLY = (204, (1, 'exactly {} option(s)'))


class CheckboxValidator(Validator):
    class Type(Validator.Type):
        UNKNOWN = (-1, None)
        DEFAULT = (7, CheckboxTypes)

    @classmethod
    def _parse_arg_list(cls, args, type_, subtype):
        try:
            return [int(arg) for arg in args], True
        except ValueError:
            return args, False

    def _validate(self, elem, values: List[List[str]]):
        required = self.args[0]
        cnt = len(values[0])
        if cast('Checkboxes', elem)._other_value is not None:
            cnt += 1
        if self.subtype is CheckboxTypes.AT_LEAST:
            is_ok = cnt >= required
        elif self.subtype is CheckboxTypes.AT_MOST:
            is_ok = cnt <= required
        elif self.subtype is CheckboxTypes.EXACTLY:
            is_ok = cnt == required
        else:
            raise NotImplementedError()
        if not is_ok:
            raise InvalidChoiceCount(elem, cnt)

    def _descr(self):
        return 'Select ' + super()._descr()
